#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running CAS server.  You must use an ~/.authinfo
#       file to specify your username and password.  The CAS host and port must
#       be specified using the CASHOST and CASPORT environment variables.
#       A specific protocol ('cas', 'http', 'https', or 'auto') can be set using
#       the CASPROTOCOL environment variable.

import os

import swat
import swat.utils.testing as tm
from dlpy.images import ImageTable

USER, PASSWD = tm.get_user_pass()
HOST, PORT, PROTOCOL = tm.get_host_port_proto()


class TestImageTable(tm.TestCase):
    # Create a class attribute to hold the cas host type
    server_type = None

    def setUp(self):
        swat.reset_option()
        swat.options.cas.print_messages = False
        swat.options.interactive_mode = False

        self.s = swat.CAS(HOST, PORT, USER, PASSWD, protocol=PROTOCOL)

        if type(self).server_type is None:
            # Set once per class and have every test use it. No need to change between tests.
            type(self).server_type = tm.get_cas_host_type(self.s)

        self.srcLib = tm.get_casout_lib(self.server_type)
        filename = os.path.join(os.path.dirname(__file__), 'datasources', 'ImageData.sashdat')
        r = tm.load_data(self.s, filename, self.server_type)

        self.tablename = r['tableName']
        self.assertNotEqual(self.tablename, None)
        self.table = ImageTable.from_table(r['casTable'])

    def tearDown(self):
        # tear down tests
        try:
            self.s.endsession()
        except swat.SWATError:
            pass
        del self.s
        swat.reset_option()

    def test_table_type(self):
        self.assertTrue(isinstance(self.table, ImageTable))

    def test_colspec_str(self):
        dtype1 = 'varbinary'
        dtype2 = 'varchar'
        # if self.s._protocol in ['http', 'https']:
        #     dtype1 = 'binary'

        out = self.table.columninfo(inputs='_image_').ColumnInfo
        column_list = ['Column', 'ID', 'Type', 'RawLength', 'FormattedLength', 'NFL', 'NFD']
        value_list = ['_image_', 1, dtype1, 858354, 858354, 0, 0]
        for i, k in zip(column_list, value_list):
            self.assertEqual(out[i][0], k)

        out = self.table.columninfo(inputs='_label_').ColumnInfo
        column_list = ['Column', 'ID', 'Type', 'RawLength', 'FormattedLength', 'NFL', 'NFD']
        value_list = ['_label_', 2, dtype2, 7, 7, 0, 0]
        for i, k in zip(column_list, value_list):
            self.assertEqual(out[i][0], k)

        out = self.table.columninfo(inputs='_filename_0').ColumnInfo
        column_list = ['Column', 'ID', 'Type', 'RawLength', 'FormattedLength', 'NFL', 'NFD']
        value_list = ['_filename_0', 3, dtype2, 17, 17, 0, 0]
        for i, k in zip(column_list, value_list):
            self.assertEqual(out[i][0], k)

    def test_image_summary(self):
        out = self.table.image_summary
        column_list = ['jpg', 'minWidth', 'maxWidth', 'minHeight', 'maxHeight', 'meanWidth',
                       'meanHeight', 'mean1stChannel', 'min1stChannel', 'max1stChannel',
                       'mean2ndChannel', 'min2ndChannel', 'max2ndChannel', 'mean3rdChannel',
                       'min3rdChannel', 'max3rdChannel']
        value_list = [21, 500, 1024, 272, 1024, 840, 813, 134, 0, 255, 124, 0, 255, 87, 0, 255]
        for i, k in zip(column_list, value_list):
            self.assertEqual(int(out[i]), k)

    def test_resize(self):
        out = self.table.resize(width=200, height=200, inplace=False)
        out = out.image_summary
        column_list = ['jpg', 'minWidth', 'maxWidth', 'minHeight', 'maxHeight', 'meanWidth',
                       'meanHeight', 'mean1stChannel', 'min1stChannel', 'max1stChannel',
                       'mean2ndChannel', 'min2ndChannel', 'max2ndChannel', 'mean3rdChannel',
                       'min3rdChannel', 'max3rdChannel']
        value_list = [21, 200, 200, 200, 200, 200, 200, 134, 0, 255, 123, 0, 255, 87, 0, 255]
        for i, k in zip(column_list, value_list):
            self.assertEqual(int(out[i]), k)

    def test_crop(self):
        out = self.table.crop(x=0, y=0, width=200, height=200, inplace=False)
        out = out.image_summary
        column_list = ['jpg', 'minWidth', 'maxWidth', 'minHeight', 'maxHeight', 'meanWidth',
                       'meanHeight', 'mean1stChannel', 'min1stChannel', 'max1stChannel',
                       'mean2ndChannel', 'min2ndChannel', 'max2ndChannel', 'mean3rdChannel',
                       'min3rdChannel', 'max3rdChannel']
        value_list = [21, 200, 200, 200, 200, 200, 200, 137, 0, 255, 123, 0, 255, 86, 0, 255]
        for i, k in zip(column_list, value_list):
            self.assertEqual(int(out[i]), k)

    def test_as_patches(self):
        out = self.table.as_patches(x=0, y=0, width=200, height=200, step_size=100,
                                    output_width=100, output_height=100, inplace=False)
        out = out.image_summary
        column_list = ['jpg', 'minWidth', 'maxWidth', 'minHeight', 'maxHeight', 'meanWidth',
                       'meanHeight', 'mean1stChannel', 'min1stChannel', 'max1stChannel',
                       'mean2ndChannel', 'min2ndChannel', 'max2ndChannel', 'mean3rdChannel',
                       'min3rdChannel', 'max3rdChannel']
        value_list = [1312, 100, 100, 100, 100, 100, 100, 129, 0, 255, 125, 0, 255, 98, 0, 255]
        for i, k in zip(column_list, value_list):
            self.assertEqual(int(out[i]), k)


if __name__ == '__main__':
    tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' Utility functions for the DLPy package '''

import os
import random
import re
import string

import matplotlib.pyplot as plt
import numpy as np
import six
import swat as sw
from swat.cas.table import CASTable


def random_name(name='ImageData', length=6):
    '''
    Generate random name

    Parameters
    ----------
    name : string, optional
        Prefix of the generated name
    length : int, optional
        Length of the random characters in the name

    Returns
    -------
    string

    '''
    return name + '_' + ''.join(random.sample(
        string.ascii_uppercase + string.ascii_lowercase + string.digits, length))


def input_table_check(input_table):
    '''
    Unify the input_table format

    Parameters
    ----------
    input_table : CASTable or string or dict
        Input table specification

    Returns
    -------
    dict
        Input table parameters

    '''
    if isinstance(input_table, six.string_types):
        input_table = dict(name=input_table)
    elif isinstance(input_table, dict):
        input_table = input_table
    elif isinstance(input_table, CASTable):
        input_table = input_table.to_table_params()
    else:
        raise TypeError('input_table must be one of the following:\n'
                        '1. A CAS table object;\n'
                        '2. A string specifies the name of the CAS table,\n'
                        '3. A dictionary specifies the CAS table\n'
                        '4. An ImageTable object.')
    return input_table


def prod_without_none(array):
    '''
    Compute the product of an iterable array with None as its element

    Parameters
    ----------
    array : iterable-of-numeric
        The numbers to use as input

    Returns
    -------
    numeric
        Product of all the elements of the array

    '''
    prod = 1
    for i in array:
        if i is not None:
            prod *= i
    return prod


def get_max_size(start_path='.'):
    '''
    Get the max size of files in a folder including sub-folders

    Parameters
    ----------
    start_path : string, optional
        The directory to start the file search

    Returns
    -------
    int

    '''
    max_size = 0
    for dirpath, dirnames, filenames in os.walk(start_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            file_size = os.path.getsize(fp)
            if file_size > max_size:
                max_size = file_size
    return max_size


def image_blocksize(width, height):
    '''
    Determine blocksize according to imagesize in the table

    Parameters
    ----------
    width : int
        The width of the image
    height : int
        The height of the image

    Returns
    -------
    int

    '''
    return width * height * 3 / 1024


def predicted_prob_barplot(ax, labels, values):
    '''
    Generate a horizontal barplot for the predict probability

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot to
    labels : list-of-strings
        Predicted class labels
    values : list-of-numeric
        Predicted probabilities

    Returns
    -------
    :class:`matplotlib.axes.Axes`

    '''
    y_pos = (0.2 + np.arange(len(labels))) / (1 + len(labels))
    width = 0.8 / (1 + len(labels))
    colors = ['blue', 'green', 'yellow', 'orange', 'red']
    for i in range(len(labels)):
        ax.barh(y_pos[i], values[i], width, align='center',
                color=colors[i], ecolor='black')
        ax.text(values[i] + 0.01, y_pos[i], '{:.2%}'.format(values[i]))
    ax.set_yticks(y_pos)
    ax.set_yticklabels(labels, rotation=45)
    ax.set_xlabel('Probability')
    ax.set_xticks([0, 0.25, 0.5, 0.75, 1, 1.1])
    ax.set_xticklabels(['0%', '25%', '50%', '75%', '100%'])
    ax.set_title('Predicted Probability')
    return ax


def plot_predict_res(image, label, labels, values):
    '''
    Generate a side by side plot of the predicted result

    Parameters
    ----------
    image :
        Specifies the orginal image to be classified.
    label : string
        Specifies the class name of the image.
    labels : list-of-strings
        Predicted class labels
    values : list-of-numeric
        Predicted probabilities

    Returns
    -------
    :class:`matplotlib.axes.Axes`

    '''
    fig = plt.figure(figsize=(12, 5))
    ax1 = fig.add_subplot(1, 2, 1)
    ax1.set_title('{}'.format(label))
    ax1.imshow(image)
    ax1.axis('off')
    ax2 = fig.add_subplot(1, 2, 2)
    predicted_prob_barplot(ax2, labels, values)


def camelcase_to_underscore(strings):
    ''' Convert camelcase to underscore '''
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', strings)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def add_caslib(conn, path):
    '''
    Add a new caslib, as needed

    Parameters
    ----------
    conn : CAS
        The CAS connection object
    path : string
        Specifies the server-side path to check

    Returns
    -------
    string
        The name of the caslib pointing to the path

    '''
    if path in conn.caslibinfo().CASLibInfo.Path.tolist():
        cas_lib_name = conn.caslibinfo().CASLibInfo[
            conn.caslibinfo().CASLibInfo.Path == path]['Name']

        return cas_lib_name.tolist()[0]
    else:
        cas_lib_name = random_name('Caslib', 6)
        conn.retrieve('table.addcaslib', message_level='error',
                      name=cas_lib_name, path=path, activeOnAdd=False,
                      dataSource=dict(srcType='DNFS'))
        return cas_lib_name


def upload_astore(conn, path, table_name=None):
    '''
    Load the local astore file to server

    Parameters
    ----------
    conn : CAS
        The CAS connection object
    path : string
        Specifies the client-side path of the astore file
    table_name : string, or casout options
        Specifies the name of the cas table on server to put the astore object

    '''
    conn.loadactionset('astore')

    with open(path, 'br') as f:
        astore_byte = f.read()

    store_ = sw.blob(astore_byte)

    if table_name is None:
        table_name = random_name('ASTORE')
    conn.astore.upload(rstore=table_name, store=store_)


def unify_keys(dic):
    '''
    Change all the key names in a dictionary to lower case, remove "_" in the key names.

    Parameters
    ----------
    dic : dict

    Returns
    -------
    dict
        dictionary with updated key names

    '''

    old_names = list(dic.keys())
    new_names = [item.lower().replace('_', '') for item in old_names]
    for new_name, old_name in zip(new_names, old_names):
        dic[new_name] = dic.pop(old_name)

    return dic


def check_caslib(conn, path):
    '''
    Check whether the specified path is in the caslibs of the current session.

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object

    path : str
        Specifies the name of the path.

    Returns
    -------
    flag : bool
        Specifies if path exist in session.
    caslib_name : str (if exist)
        Specifies the name of the caslib that contain the path.

    '''
    paths = conn.caslibinfo().CASLibInfo.Path.tolist()
    caslibs = conn.caslibinfo().CASLibInfo.Name.tolist()

    if path in paths:
        caslibname = caslibs[paths.index(path)]
        return True, caslibname
    else:
        return False



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' Write python codes for create sas model '''

import sys


# model/input layer definition
def write_input_layer(model_name='sas', layer_name='data', channels='-1',
                      width='-1', height='-1', scale='1.0'):
    '''
    Generate Python code defining a SAS deep learning input layer

    Parameters
    ----------
    model_name : string
       Name for deep learning model
    layer_name : string
       Layer name
    channels : string
       number of input channels
    width : string
       image width
    height : string
       image height
    scale : string
       scaling factor to apply to raw image pixel data

    Returns
    -------
    string
        String representing Python code defining a SAS deep learning input layer

    '''
    out = [
        'import sys',
        '',
        'def sas_model_gen(s, input_crop_type=None, input_channel_offset=None, input_image_size=None):',
        '   # quick check for deeplearn actionset',
        '   actionset_list = s.actionsetinfo().setinfo.actionset.tolist()',
        '   actionset_list = [item.lower() for item in actionset_list]',
        '   if "deeplearn" not in actionset_list:s.loadactionset("deeplearn")',
        '   ',
        '   # quick error-checking and default setting',
        '   if (input_crop_type is None):',
        '       input_crop_type="NONE"',
        '   else:',
        '       if (input_crop_type.upper() != "NONE") and (input_crop_type.upper() != "UNIQUE"):',
        '           sys.exit("ERROR: input_crop_type can only be NONE or UNIQUE")',
        '',
        '   if (input_image_size is not None):',
        '       channels = input_image_size[0]',
        '       if (len(input_image_size) == 2):',
        '           height = width = input_image_size[1]',
        '       elif (len(inputImageSize) == 3):',
        '           height,width = input_image_size[1:]',
        '       else:',
        '           sys.exit("ERROR: input_image_size must be a tuple with two or three entries")',
        '',
        '   # instantiate model',
        '   s.buildModel(model=dict(name=' + repr(model_name) + ',replace=True),type="CNN")',
        '',
        '   # input layer',
        '   nchannels=' + channels,
        '   if input_channel_offset is None and nchannels==3:',
        '       print("INFO: setting channel mean values to ImageNet means")',
        '       input_channel_offset = [103.939, 116.779, 123.68]',
        '       s.addLayer(model=' + repr(model_name) + ', name=' + repr(layer_name) + ',',
        '                  layer=dict( type="input", nchannels=' + channels + ', width=' + width + ', height=' + height + ',',
        '                  scale = ' + scale + ', randomcrop=input_crop_type, offsets=input_channel_offset))',
        '   elif input_channel_offset is not None:',
        '       s.addLayer(model=' + repr(model_name) + ', name=' + repr(layer_name) + ',',
        '                  layer=dict( type="input", nchannels=' + channels + ', width=' + width + ', height=' + height + ',',
        '                  scale = ' + scale + ', randomcrop=input_crop_type, offsets=input_channel_offset))',
        '   else:',
        '       s.addLayer(model=' + repr(model_name) + ', name=' + repr(layer_name) + ',',
        '                  layer=dict( type="input", nchannels=' + channels + ', width=' + width + ', height=' + height + ',',
        '                  scale = ' + scale + ', randomcrop=input_crop_type))'
    ]
    return '\n'.join(out)


# convolution layer definition
def write_convolution_layer(model_name='sas', layer_name='conv', nfilters='-1',
                            width='3', height='3', stride='1', nobias='False',
                            activation='identity', dropout='0', src_layer='none'):
    '''
    Generate Python code defining a SAS deep learning convolution layer

    Parameters
    ----------
    model_name : string, optional
       Name for deep learning model
    layer_name : string, optional
       Layer name
    nfilters : string, optional
       number of output feature maps
    width : string, optional
       image width
    height : string, optional
       image height
    stride : string, optional
       vertical/horizontal step size in pixels
    nobias : string, optional
       omit (True) or retain (False) the bias term
    activation : string, optional
       activation function
    dropout : string, optional
       dropout factor (0 < dropout < 1.0)
    src_layer : string, optional
       source layer(s) for the convolution layer

    Returns
    -------
    string

    '''
    out = [
        '   s.addLayer(model=' + repr(model_name) + ', name=' + repr(layer_name) + ',',
        '              layer=dict(type="convolution", nfilters=' + nfilters + ', width=' + width + ', height=' + height + ',',
        '                         stride=' + stride + ', nobias=' + nobias + ', act=' + repr(
            activation) + ', dropout=' + dropout + '), \n',
        '              srcLayers=' + src_layer + ')'
    ]
    return '\n'.join(out)


# batch normalization layer definition
def write_batch_norm_layer(model_name='sas', layer_name='bn',
                           activation='identity', src_layer='none'):
    '''
    Generate Python code defining a SAS deep learning batch normalization layer

    Parameters
    ----------
    model_name : string, optional
       Name for deep learning model
    layer_name : string, optional
       Layer name
    activation : string, optional
       activation function
    src_layer : string, optional
       source layer(s) for the convolution layer

    Returns
    -------
    string

    '''
    out = [
        '   s.addLayer(model=' + repr(model_name) + ', name=' + repr(layer_name) + ',',
        '              layer=dict( type="batchnorm", act=' + repr(activation) + '),',
        '              srcLayers=' + src_layer + ')'
    ]
    return '\n'.join(out)


# pooling layer definition
def write_pooling_layer(model_name='sas', layer_name='pool',
                        width='2', height='2', stride='2', type='max',
                        dropout='0', src_layer='none'):
    '''
    Generate Python code defining a SAS deep learning pooling layer

    Parameters
    ----------
    model_name : string, optional
       Name for deep learning model
    layer_name : string, optional
       Layer name
    width : string, optional
       image width
    height : string, optional
       image height
    stride : string, optional
       vertical/horizontal step size in pixels
    type : string, optional
       pooling type
    dropout : string, optional
       dropout factor (0 < dropout < 1.0)
    src_layer : string, optional
       source layer(s) for the convolution layer

    Returns
    -------
    string

    '''
    out = [
        '   s.addLayer(model=' + repr(model_name) + ', name=' + repr(layer_name) + ',',
        '              layer=dict(type="pooling", width=' + width + ', height=' + height + ',',
        '                         stride=' + stride + ', pool=' + repr(type) + ', dropout=' + dropout + '),',
        '              srcLayers=' + src_layer + ')'
    ]
    return '\n'.join(out)


# residual layer definition
def write_residual_layer(model_name='sas', layer_name='residual',
                         activation='identity', src_layer='none'):
    '''
    Generate Python code defining a SAS deep learning residual layer

    Parameters
    ----------
    model_name : string, optional
       Name for deep learning model
    layer_name : string, optional
       Layer name
    activation : string, optional
       activation function
    src_layer : string, optional
       source layer(s) for the convolution layer

    Returns
    -------
    string

    '''
    out = [
        '   s.addLayer(model=' + repr(model_name) + ', name=' + repr(layer_name) + ',',
        '              layer=dict( type="residual", act="' + activation + '"),',
        '              srcLayers=' + src_layer + ')'
    ]
    return '\n'.join(out)


# fully connected layer definition
def write_full_connect_layer(model_name='sas', layer_name='fullconnect',
                             nrof_neurons='-1', nobias='true',
                             activation='identity', type='fullconnect', dropout='0',
                             src_layer='none'):
    '''
    Generate Python code defining a SAS deep learning fully connected layer

    Parameters
    ----------
    model_name : string, optional
       Name for deep learning model
    layer_name : string, optional
       Layer name
    nrof_neurons : string, optional
       number of output neurons
    nobias : string, optional
       omit (True) or retain (False) the bias term
    activation : string, optional
       activation function
    type : string, optional
       fully connected layer type (fullconnect or output)
    dropout : string, optional
       dropout factor (0 < dropout < 1.0)
    src_layer : string, optional
       source layer(s) for the convolution layer

    Returns
    -------
    string

    '''
    if (type == 'fullconnect'):
        out = [
            '   s.addLayer(model=' + repr(model_name) + ', name=' + repr(layer_name) + ',',
            '              layer=dict(type=' + repr(type) + ', n=' + nrof_neurons + ',',
            '                         nobias=' + nobias + ', act=' + repr(activation) + ', dropout=' + dropout + '),',
            '              srcLayers=' + src_layer + ')'
        ]
    else:
        out = [
            '   s.addLayer(model=' + repr(model_name) + ', name=' + repr(layer_name) + ',',
            '              layer=dict(type=' + repr(type) + ', n=' + nrof_neurons + ',',
            '                         nobias=' + nobias + ', act=' + repr(activation) + '),',
            '              srcLayers=' + src_layer + ')'
        ]
    return '\n'.join(out)


# Python __main__ function
def write_main_entry(model_name):
    '''
    Generate Python code defining the __main__ Python entry point

    Parameters
    ----------
    model_name : string
       Name for deep learning model

    Returns
    -------
    string

    '''
    out = [
        '#############################################################',
        'if __name__ == "__main__":',
        '   sys.exit("ERROR: this module defines the ' + model_name + ' model.  Do not call directly.")'
    ]
    return '\n'.join(out)


#############################################################
if __name__ == '__main__':
    sys.exit('ERROR: this module cannot be invoked from the command line.')



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from ..utils import input_table_check


def ResNet50_Model(s, model_table='RESNET50', n_channels=3, width=224, height=224,
                   random_crop=None, offsets=None):
    '''
    ResNet50 model definition

    Parameters
    ----------
    s : CAS
        Specifies the CAS connection object.
    model_table : string, dict or CAS table, optional
        Specifies the CAS table to store the model.
    n_channels : int, optional
        Specifies the number of the channels of the input layer.
        Default: 3
    width : int, optional
        Specifies the width of the input layer.
        Default: 224
    height : int, optional
        Specifies the height of the input layer.
        Default: 224
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters.deepLearn. Only the images with one or both dimensions
        that are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default	: 'unique'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final
        input data is set after applying scaling and subtracting the
        specified offsets.deepLearn.
        Default: (103.939, 116.779, 123.68)

    Returns
    -------
    None
        CAS table defining the model is created

    '''

    model_table_opts = input_table_check(model_table)

    if random_crop is None:
        random_crop = 'none'
    elif random_crop.lower() not in ['none', 'unique']:
        raise ValueError('random_crop can only be "none" or "unique"')

    if offsets is None:
        offsets = [103.939, 116.779, 123.68]

    # instantiate model
    s.deepLearn.buildModel(model=dict(replace=True, **model_table_opts), type='CNN')

    # input layer
    s.deepLearn.addLayer(model=model_table_opts, name='data',
                         layer=dict(type='input', nchannels=n_channels, width=width, height=height,
                                    randomcrop=random_crop, offsets=offsets))

    # -------------------- Layer 1 ----------------------

    # conv1 layer: 64 channels, 7x7 conv, stride=2; output = 112 x 112 */
    s.deepLearn.addLayer(model=model_table_opts, name='conv1',
                         layer=dict(type='convolution', nFilters=64, width=7, height=7,
                                    stride=2, act='identity'),
                         srcLayers=['data'])

    # conv1 batch norm layer: 64 channels, output = 112 x 112 */
    s.deepLearn.addLayer(model=model_table_opts, name='bn_conv1',
                         layer=dict(type='batchnorm', act='relu'), srcLayers=['conv1'])

    # pool1 layer: 64 channels, 3x3 pooling, output = 56 x 56 */
    s.deepLearn.addLayer(model=model_table_opts, name='pool1',
                         layer=dict(type='pooling', width=3, height=3, stride=2, pool='max'),
                         srcLayers=['bn_conv1'])

    # ------------------- Residual Layer 2A -----------------------

    # res2a_branch1 layer: 256 channels, 1x1 conv, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='res2a_branch1',
                         layer=dict(type='convolution', nFilters=256, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['pool1'])

    # res2a_branch1 batch norm layer: 256 channels, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='bn2a_branch1',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res2a_branch1'])

    # res2a_branch2a layer: 64 channels, 1x1 conv, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='res2a_branch2a',
                         layer=dict(type='convolution', nFilters=64, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['pool1'])

    # res2a_branch2a batch norm layer: 64 channels, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='bn2a_branch2a',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res2a_branch2a'])

    # res2a_branch2b layer: 64 channels, 3x3 conv, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='res2a_branch2b',
                         layer=dict(type='convolution', nFilters=64, width=3, height=3,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn2a_branch2a'])

    # res2a_branch2b batch norm layer: 64 channels, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='bn2a_branch2b',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res2a_branch2b'])

    # res2a_branch2c layer: 256 channels, 1x1 conv, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='res2a_branch2c',
                         layer=dict(type='convolution', nFilters=256, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn2a_branch2b'])

    # res2a_branch2c batch norm layer: 256 channels, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='bn2a_branch2c',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res2a_branch2c'])

    # res2a residual layer: 256 channels, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='res2a',
                         layer=dict(type='residual', act='relu'),
                         srcLayers=['bn2a_branch2c', 'bn2a_branch1'])

    # ------------------- Residual Layer 2B -----------------------

    # res2b_branch2a layer: 64 channels, 1x1 conv, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='res2b_branch2a',
                         layer=dict(type='convolution', nFilters=64, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['res2a'])

    # res2b_branch2a batch norm layer: 64 channels, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='bn2b_branch2a',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res2b_branch2a'])

    # res2b_branch2b layer: 64 channels, 3x3 conv, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='res2b_branch2b',
                         layer=dict(type='convolution', nFilters=64, width=3, height=3,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn2b_branch2a'])

    # res2b_branch2b batch norm layer: 64 channels, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='bn2b_branch2b',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res2b_branch2b'])

    # res2b_branch2c layer: 256 channels, 1x1 conv, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='res2b_branch2c',
                         layer=dict(type='convolution', nFilters=256, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn2b_branch2b'])

    # res2b_branch2c batch norm layer: 256 channels, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='bn2b_branch2c',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res2b_branch2c'])

    # res2b residual layer: 256 channels, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='res2b',
                         layer=dict(type='residual', act='relu'),
                         srcLayers=['bn2b_branch2c', 'res2a'])

    # ------------------- Residual Layer 2C -----------------------

    # res2c_branch2a layer: 64 channels, 1x1 conv, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='res2c_branch2a',
                         layer=dict(type='convolution', nFilters=64, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['res2b'])

    # res2c_branch2a batch norm layer: 64 channels, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='bn2c_branch2a',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res2c_branch2a'])

    # res2c_branch2b layer: 64 channels, 3x3 conv, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='res2c_branch2b',
                         layer=dict(type='convolution', nFilters=64, width=3, height=3,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn2c_branch2a'])

    # res2c_branch2b batch norm layer: 64 channels, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='bn2c_branch2b',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res2c_branch2b'])

    # res2c_branch2c layer: 256 channels, 1x1 conv, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='res2c_branch2c',
                         layer=dict(type='convolution', nFilters=256, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn2c_branch2b'])

    # res2c_branch2c batch norm layer: 256 channels, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='bn2c_branch2c',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res2c_branch2c'])

    # res2c residual layer: 256 channels, output = 56 x 56
    s.deepLearn.addLayer(model=model_table_opts, name='res2c',
                         layer=dict(type='residual', act='relu'),
                         srcLayers=['bn2c_branch2c', 'res2b'])

    # ------------- Layer 3A --------------------

    # res3a_branch1 layer: 512 channels, 1x1 conv, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='res3a_branch1',
                         layer=dict(type='convolution', nFilters=512, width=1, height=1,
                                    stride=2, includebias=False, act='identity'),
                         srcLayers=['res2c'])

    # res3a_branch1 batch norm layer: 512 channels, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='bn3a_branch1',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res3a_branch1'])

    # res3a_branch2a layer: 128 channels, 1x1 conv, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='res3a_branch2a',
                         layer=dict(type='convolution', nFilters=128, width=1, height=1,
                                    stride=2, includebias=False, act='identity'),
                         srcLayers=['res2c'])

    # res3a_branch2a batch norm layer: 128 channels, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='bn3a_branch2a',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res3a_branch2a'])

    # res3a_branch2b layer: 128 channels, 3x3 conv, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='res3a_branch2b',
                         layer=dict(type='convolution', nFilters=128, width=3, height=3,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn3a_branch2a'])

    # res3a_branch2b batch norm layer: 128 channels, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='bn3a_branch2b',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res3a_branch2b'])

    # res3a_branch2c layer: 512 channels, 1x1 conv, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='res3a_branch2c',
                         layer=dict(type='convolution', nFilters=512, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn3a_branch2b'])

    # res3a_branch2c batch norm layer: 512 channels, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='bn3a_branch2c',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res3a_branch2c'])

    # res3a residual layer: 512 channels, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='res3a',
                         layer=dict(type='residual', act='relu'),
                         srcLayers=['bn3a_branch2c', 'bn3a_branch1'])

    # ------------------- Residual Layer 3B -----------------------

    # res3b_branch2a layer: 128 channels, 1x1 conv, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='res3b_branch2a',
                         layer=dict(type='convolution', nFilters=128, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['res3a'])

    # res3b_branch2a batch norm layer: 128 channels, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='bn3b_branch2a',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res3b_branch2a'])

    # res3b_branch2b layer: 128 channels, 3x3 conv, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='res3b_branch2b',
                         layer=dict(type='convolution', nFilters=128, width=3, height=3,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn3b_branch2a'])

    # res3b_branch2b batch norm layer: 128 channels, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='bn3b_branch2b',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res3b_branch2b'])

    # res3b_branch2c layer: 512 channels, 1x1 conv, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='res3b_branch2c',
                         layer=dict(type='convolution', nFilters=512, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn3b_branch2b'])

    # res3b_branch2c batch norm layer: 512 channels, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='bn3b_branch2c',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res3b_branch2c'])

    # res3b residual layer: 512 channels, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='res3b',
                         layer=dict(type='residual', act='relu'),
                         srcLayers=['bn3b_branch2c', 'res3a'])

    # ------------------- Residual Layer 3C -----------------------

    # res3c_branch2a layer: 128 channels, 1x1 conv, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='res3c_branch2a',
                         layer=dict(type='convolution', nFilters=128, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['res3b'])

    # res3c_branch2a batch norm layer: 128 channels, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='bn3c_branch2a',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res3c_branch2a'])

    # res3c_branch2b layer: 128 channels, 3x3 conv, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='res3c_branch2b',
                         layer=dict(type='convolution', nFilters=128, width=3, height=3,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn3c_branch2a'])

    # res3c_branch2b batch norm layer: 128 channels, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='bn3c_branch2b',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res3c_branch2b'])

    # res3c_branch2c layer: 512 channels, 1x1 conv, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='res3c_branch2c',
                         layer=dict(type='convolution', nFilters=512, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn3c_branch2b'])

    # res3c_branch2c batch norm layer: 512 channels, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='bn3c_branch2c',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res3c_branch2c'])

    # res3c residual layer: 512 channels, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='res3c',
                         layer=dict(type='residual', act='relu'),
                         srcLayers=['bn3c_branch2c', 'res3b'])

    # ------------------- Residual Layer 3D -----------------------

    # res3d_branch2a layer: 128 channels, 1x1 conv, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='res3d_branch2a',
                         layer=dict(type='convolution', nFilters=128, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['res3c'])

    # res3d_branch2a batch norm layer: 128 channels, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='bn3d_branch2a',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res3d_branch2a'])

    # res3d_branch2b layer: 128 channels, 3x3 conv, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='res3d_branch2b',
                         layer=dict(type='convolution', nFilters=128, width=3, height=3,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn3d_branch2a'])

    # res3d_branch2b batch norm layer: 128 channels, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='bn3d_branch2b',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res3d_branch2b'])

    # res3d_branch2c layer: 512 channels, 1x1 conv, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='res3d_branch2c',
                         layer=dict(type='convolution', nFilters=512, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn3d_branch2b'])

    # res3d_branch2c batch norm layer: 512 channels, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='bn3d_branch2c',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res3d_branch2c'])

    # res3d residual layer: 512 channels, output = 28 x 28
    s.deepLearn.addLayer(model=model_table_opts, name='res3d',
                         layer=dict(type='residual', act='relu'),
                         srcLayers=['bn3d_branch2c', 'res3c'])

    # ------------- Layer 4A --------------------

    # res4a_branch1 layer: 1024 channels, 1x1 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4a_branch1',
                         layer=dict(type='convolution', nFilters=1024, width=1, height=1,
                                    stride=2, includebias=False, act='identity'),
                         srcLayers=['res3d'])

    # res4a_branch1 batch norm layer: 1024 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4a_branch1',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res4a_branch1'])

    # res4a_branch2a layer: 256 channels, 1x1 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4a_branch2a',
                         layer=dict(type='convolution', nFilters=256, width=1, height=1,
                                    stride=2, includebias=False, act='identity'),
                         srcLayers=['res3d'])

    # res4a_branch2a batch norm layer: 256 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4a_branch2a',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res4a_branch2a'])

    # res4a_branch2b layer: 256 channels, 3x3 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4a_branch2b',
                         layer=dict(type='convolution', nFilters=256, width=3, height=3,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn4a_branch2a'])

    # res4a_branch2b batch norm layer: 256 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4a_branch2b',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res4a_branch2b'])

    # res4a_branch2c layer: 1024 channels, 1x1 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4a_branch2c',
                         layer=dict(type='convolution', nFilters=1024, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn4a_branch2b'])

    # res4a_branch2c batch norm layer: 1024 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4a_branch2c',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res4a_branch2c'])

    # res4a residual layer: 1024 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4a',
                         layer=dict(type='residual', act='relu'),
                         srcLayers=['bn4a_branch2c', 'bn4a_branch1'])

    # ------------------- Residual Layer 4B -----------------------

    # res4b_branch2a layer: 256 channels, 1x1 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4b_branch2a',
                         layer=dict(type='convolution', nFilters=256, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['res4a'])

    # res4b_branch2a batch norm layer: 256 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4b_branch2a',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res4b_branch2a'])

    # res4b_branch2b layer: 256 channels, 3x3 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4b_branch2b',
                         layer=dict(type='convolution', nFilters=256, width=3, height=3,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn4b_branch2a'])

    # res4b_branch2b batch norm layer: 256 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4b_branch2b',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res4b_branch2b'])

    # res4b_branch2c layer: 1024 channels, 1x1 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4b_branch2c',
                         layer=dict(type='convolution', nFilters=1024, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn4b_branch2b'])

    # res4b_branch2c batch norm layer: 1024 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4b_branch2c',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res4b_branch2c'])

    # res4b residual layer: 1024 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4b',
                         layer=dict(type='residual', act='relu'),
                         srcLayers=['bn4b_branch2c', 'res4a'])

    # ------------------- Residual Layer 4C -----------------------

    # res4c_branch2a layer: 256 channels, 1x1 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4c_branch2a',
                         layer=dict(type='convolution', nFilters=256, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['res4b'])

    # res4c_branch2a batch norm layer: 256 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4c_branch2a',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res4c_branch2a'])

    # res4c_branch2b layer: 256 channels, 3x3 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4c_branch2b',
                         layer=dict(type='convolution', nFilters=256, width=3, height=3,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn4c_branch2a'])

    # res4c_branch2b batch norm layer: 256 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4c_branch2b',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res4c_branch2b'])

    # res4c_branch2c layer: 1024 channels, 1x1 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4c_branch2c',
                         layer=dict(type='convolution', nFilters=1024, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn4c_branch2b'])

    # res4c_branch2c batch norm layer: 1024 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4c_branch2c',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res4c_branch2c'])

    # res4c residual layer: 1024 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4c',
                         layer=dict(type='residual', act='relu'),
                         srcLayers=['bn4c_branch2c', 'res4b'])

    # ------------------- Residual Layer 4D -----------------------

    # res4d_branch2a layer: 256 channels, 1x1 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4d_branch2a',
                         layer=dict(type='convolution', nFilters=256, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['res4c'])

    # res4d_branch2a batch norm layer: 256 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4d_branch2a',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res4d_branch2a'])

    # res4d_branch2b layer: 256 channels, 3x3 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4d_branch2b',
                         layer=dict(type='convolution', nFilters=256, width=3, height=3,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn4d_branch2a'])

    # res4d_branch2b batch norm layer: 256 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4d_branch2b',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res4d_branch2b'])

    # res4d_branch2c layer: 1024 channels, 1x1 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4d_branch2c',
                         layer=dict(type='convolution', nFilters=1024, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn4d_branch2b'])

    # res4d_branch2c batch norm layer: 1024 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4d_branch2c',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res4d_branch2c'])

    # res4d residual layer: 1024 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4d',
                         layer=dict(type='residual', act='relu'),
                         srcLayers=['bn4d_branch2c', 'res4c'])

    # ------------------- Residual Layer 4E ----------------------- */

    # res4e_branch2a layer: 256 channels, 1x1 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4e_branch2a',
                         layer=dict(type='convolution', nFilters=256, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['res4d'])

    # res4e_branch2a batch norm layer: 256 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4e_branch2a',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res4e_branch2a'])

    # res4e_branch2b layer: 256 channels, 3x3 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4e_branch2b',
                         layer=dict(type='convolution', nFilters=256, width=3, height=3,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn4e_branch2a'])

    # res4e_branch2b batch norm layer: 256 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4e_branch2b',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res4e_branch2b'])

    # res4e_branch2c layer: 1024 channels, 1x1 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4e_branch2c',
                         layer=dict(type='convolution', nFilters=1024, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn4e_branch2b'])

    # res4e_branch2c batch norm layer: 1024 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4e_branch2c',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res4e_branch2c'])

    # res4e residual layer: 1024 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4e',
                         layer=dict(type='residual', act='relu'),
                         srcLayers=['bn4e_branch2c', 'res4d'])

    # ------------------- Residual Layer 4F -----------------------

    # res4f_branch2a layer: 256 channels, 1x1 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4f_branch2a',
                         layer=dict(type='convolution', nFilters=256, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['res4e'])

    # res4f_branch2a batch norm layer: 256 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4f_branch2a',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res4f_branch2a'])

    # res4f_branch2b layer: 256 channels, 3x3 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4f_branch2b',
                         layer=dict(type='convolution', nFilters=256, width=3, height=3,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn4f_branch2a'])

    # res4f_branch2b batch norm layer: 256 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4f_branch2b',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res4f_branch2b'])

    # res4f_branch2c layer: 1024 channels, 1x1 conv, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4f_branch2c',
                         layer=dict(type='convolution', nFilters=1024, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn4f_branch2b'])

    # res4f_branch2c batch norm layer: 1024 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='bn4f_branch2c',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res4f_branch2c'])

    # res4f residual layer: 1024 channels, output = 14 x 14
    s.deepLearn.addLayer(model=model_table_opts, name='res4f',
                         layer=dict(type='residual', act='relu'),
                         srcLayers=['bn4f_branch2c', 'res4e'])

    # ------------- Layer 5A -------------------- */

    # res5a_branch1 layer: 2048 channels, 1x1 conv, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='res5a_branch1',
                         layer=dict(type='convolution', nFilters=2048, width=1, height=1,
                                    stride=2, includebias=False, act='identity'),
                         srcLayers=['res4f'])

    # res5a_branch1 batch norm layer: 2048 channels, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='bn5a_branch1',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res5a_branch1'])

    # res5a_branch2a layer: 512 channels, 1x1 conv, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='res5a_branch2a',
                         layer=dict(type='convolution', nFilters=512, width=1, height=1,
                                    stride=2, includebias=False, act='identity'),
                         srcLayers=['res4f'])

    # res5a_branch2a batch norm layer: 512 channels, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='bn5a_branch2a',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res5a_branch2a'])

    # res5a_branch2b layer: 512 channels, 3x3 conv, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='res5a_branch2b',
                         layer=dict(type='convolution', nFilters=512, width=3, height=3,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn5a_branch2a'])

    # res5a_branch2b batch norm layer: 512 channels, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='bn5a_branch2b',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res5a_branch2b'])

    # res5a_branch2c layer: 2048 channels, 1x1 conv, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='res5a_branch2c',
                         layer=dict(type='convolution', nFilters=2048, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn5a_branch2b'])

    # res5a_branch2c batch norm layer: 2048 channels, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='bn5a_branch2c',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res5a_branch2c'])

    # res5a residual layer: 2048 channels, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='res5a',
                         layer=dict(type='residual', act='relu'),
                         srcLayers=['bn5a_branch2c', 'bn5a_branch1'])

    # ------------------- Residual Layer 5B -----------------------

    # res5b_branch2a layer: 512 channels, 1x1 conv, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='res5b_branch2a',
                         layer=dict(type='convolution', nFilters=512, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['res5a'])

    # res5b_branch2a batch norm layer: 512 channels, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='bn5b_branch2a',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res5b_branch2a'])

    # res5b_branch2b layer: 512 channels, 3x3 conv, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='res5b_branch2b',
                         layer=dict(type='convolution', nFilters=512, width=3, height=3,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn5b_branch2a'])

    # res5b_branch2b batch norm layer: 512 channels, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='bn5b_branch2b',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res5b_branch2b'])

    # res5b_branch2c layer: 2048 channels, 1x1 conv, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='res5b_branch2c',
                         layer=dict(type='convolution', nFilters=2048, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn5b_branch2b'])

    # res5b_branch2c batch norm layer: 2048 channels, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='bn5b_branch2c',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res5b_branch2c'])

    # res5b residual layer: 2048 channels, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='res5b',
                         layer=dict(type='residual', act='relu'),
                         srcLayers=['bn5b_branch2c', 'res5a'])

    # ------------------- Residual Layer 5C -----------------------

    # res5c_branch2a layer: 512 channels, 1x1 conv, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='res5c_branch2a',
                         layer=dict(type='convolution', nFilters=512, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['res5b'])

    # res5c_branch2a batch norm layer: 512 channels, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='bn5c_branch2a',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res5c_branch2a'])

    # res5c_branch2b layer: 512 channels, 3x3 conv, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='res5c_branch2b',
                         layer=dict(type='convolution', nFilters=512, width=3, height=3,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn5c_branch2a'])

    # res5c_branch2b batch norm layer: 512 channels, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='bn5c_branch2b',
                         layer=dict(type='batchnorm', act='relu'),
                         srcLayers=['res5c_branch2b'])

    # res5c_branch2c layer: 2048 channels, 1x1 conv, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='res5c_branch2c',
                         layer=dict(type='convolution', nFilters=2048, width=1, height=1,
                                    stride=1, includebias=False, act='identity'),
                         srcLayers=['bn5c_branch2b'])

    # res5c_branch2c batch norm layer: 2048 channels, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='bn5c_branch2c',
                         layer=dict(type='batchnorm', act='identity'),
                         srcLayers=['res5c_branch2c'])

    # res5c residual layer: 2048 channels, output = 7 x 7
    s.deepLearn.addLayer(model=model_table_opts, name='res5c',
                         layer=dict(type='residual', act='relu'),
                         srcLayers=['bn5c_branch2c', 'res5b'])

    # ------------------- final layers ----------------------

    # pool5 layer: 2048 channels, 7x7 pooling, output = 1 x 1
    kernel_width = width // 2 // 2 // 2 // 2 // 2
    kernel_height = height // 2 // 2 // 2 // 2 // 2
    stride = kernel_width

    s.deepLearn.addLayer(model=model_table_opts, name='pool5',
                         layer=dict(type='pooling', width=kernel_width,
                                    height=kernel_height, stride=stride, pool='mean'),
                         srcLayers=['res5c'])

    # fc1000 output layer: 1000 neurons */
    s.deepLearn.addLayer(model=model_table_opts, name='fc1000',
                         layer=dict(type='output', n=1000, act='softmax'),
                         srcLayers=['pool5'])

    return s.CASTable(**model_table_opts)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' Sequential object for deep learning '''

from .model import Model


class Sequential(Model):
    '''
    Model for sequentially building of deep learning models

    Parameters
    ----------
    conn : CAS
        Specifies the CAS connection object
    layers : list of Layer objects.
        Specifies the layers of the sequential model.
    model_table : string, dict or CAS table, optional
        Specifies the CAS table to store the deep learning model.
        Default : None

    Returns
    -------
    :class:`Sequential`

    '''

    def __init__(self, conn, layers=None, model_table=None):
        Model.__init__(self, conn, model_table=model_table)

        if layers is None:
            self.layers = []
        elif type(layers) is not dict:
            raise TypeError('layers has to be a list of layer(s).')
        else:
            self.layers = layers
            if layers[-1]['type'] == 'output':
                self.compile()

    def add(self, layer):
        '''
        Add layer(s) to model

        Parameters
        ----------
        layer : Layer or list-of-Layer
            Specifies the layer to be added

        '''
        if self.layers == [] and layer.config['type'].lower() != 'input':
            raise ValueError('The first layer of the model must be an input layer')
        if len(self.layers) > 0 and layer.config['type'] is 'input':
            raise ValueError('Only the first layer of the Sequential model '
                             'can be an input layer')
        self.layers.append(layer)

        if layer.config['type'].lower() == 'input':
            print('NOTE: Input layer added.')

        elif layer.config['type'].lower() in ('convo', 'convolution'):
            print('NOTE: Convolutional layer added.')

        elif layer.config['type'].lower() in ('pool', 'pooling'):
            print('NOTE: Pooling layer added.')

        elif layer.config['type'].lower() in ('fc', 'fullconnect'):
            print('NOTE: Fully-connected layer added.')

        elif layer.config['type'].lower() == 'batchnorm':
            print('NOTE: Batch Normalization layer added.')

        elif layer.config['type'].lower() == 'block':
            print('NOTE: A block of layers added.')

        elif layer.config['type'].lower() == 'output':
            print('NOTE: Output layer added.')
            self.compile()

    def pop(self, loc=-1):
        '''
        Delete layer(s) from model and return it

        Parameters
        ----------
        loc : int
            Specifies the index of the layer in the model

        '''
        if len(self.layers) > 0:
            return self.layers.pop(loc)

    def switch(self, loc1, loc2):
        '''
        Switch the oder of two layers in the model.

        Parameters
        ----------
        loc1 : int
            Specifies the index of the first layer
        loc2 : int
            Specifies the index of the second layer

        '''
        self.layers[loc1], self.layers[loc2] = self.layers[loc2], self.layers[loc1]

    def compile(self):
        '''
        Convert the layer objects into Viya options

        '''
        if self.layers[0].config['type'] != 'input':
            raise ValueError('The first layer of the model must be an input layer')
        # if self.layers[-1].config['type'] != 'output':
        #     raise ValueError('The last layer of the model must be an output layer')
        self._retrieve_('deeplearn.buildmodel', model=dict(
            name=self.model_name, replace=True), type='CNN')

        conv_num = 1
        fc_num = 1
        bn_num = 1
        block_num = 1

        compiled_layers = []

        for layer in self.layers:
            if layer.config['type'] == 'block':
                options = layer.compile(src_layer=output_layer, block_num=block_num)
                block_num += 1
                for item in layer.layers:
                    compiled_layers.append(item)
                output_layer = layer.layers[-1]
                for option in options:
                    self._retrieve_('deeplearn.addlayer', model=self.model_name, **option)
            else:
                # Name each layer of the model.
                if layer.config['type'] == 'input':
                    if layer.name is None:
                        layer.name = 'Data'
                else:
                    layer.src_layers = [output_layer]
                    if layer.config['type'].lower() in ('convo', 'convolution'):
                        if layer.name is None:
                            layer.name = 'Conv{}_{}'.format(block_num, conv_num)
                            conv_num += 1
                    elif layer.config['type'].lower() == 'batchnorm':
                        if layer.name is None:
                            layer.name = 'BN{}_{}'.format(block_num, bn_num)
                            bn_num += 1
                    elif layer.config['type'].lower() in ('pool', 'pooling'):
                        if layer.name is None:
                            layer.name = 'Pool{}'.format(block_num)
                            block_num += 1
                            conv_num = 1
                    elif layer.config['type'].lower() in ('fc', 'fullconnect'):
                        if layer.name is None:
                            layer.name = 'FC{}'.format(fc_num)
                            fc_num += 1
                    elif layer.config['type'].lower() == 'output':
                        if layer.name is None:
                            layer.name = 'Output'
                    else:
                        raise ValueError('{} is not a supported layer type'.format(
                            layer.config['type']))

                option = layer.to_model_params()
                compiled_layers.append(layer)
                output_layer = layer

                self._retrieve_('deeplearn.addlayer', model=self.model_name, **option)

        print('NOTE: Model compiled successfully.')
        self.layers = compiled_layers



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' Supporting functions for keras model conversion '''

import sys
import h5py
import numpy as np
from keras import backend as K
from keras.engine.topology import preprocess_weights_for_loading


# let Keras read parameters and then transform to format needed for SAS deep learning
# NOTE: modified version of Keras function load_weights_from_hdf5_group()
def write_keras_hdf5_from_file(model, hdf5_in, hdf5_out):
    '''
    Generate an HDF5 file with trained model parameters given a Keras definition

    Parameters
    ----------
    model : Keras model
       Keras deep learning model
    hdf5_in : string
       Fully qualified file name of Keras HDF5 file
    hdf5_out : string
       Fully qualified file name of SAS-compatible HDF5 file

    '''
    # open input/output files
    try:
        f_in = h5py.File(hdf5_in, 'r')
    except IOError:
        sys.exit('File ' + hdf5_in + ' does not exist')

    try:
        f_out = h5py.File(hdf5_out, 'w')
    except IOError:
        sys.exit('File ' + hdf5_out + ' could not be created')

    if 'keras_version' in f_in.attrs:
        original_keras_version = f_in.attrs['keras_version'].decode('utf8')
    else:
        original_keras_version = '1'
    if 'backend' in f_in.attrs:
        original_backend = f_in.attrs['backend'].decode('utf8')
    else:
        original_backend = None

    try:

        image_data_format = K.image_data_format()

        # determine layers with weights
        filtered_layers = []
        for layer in model.layers:
            weights = layer.weights
            if weights:
                filtered_layers.append(layer)

        layer_names = [n.decode('utf8') for n in f_in.attrs['layer_names']]
        filtered_layer_names = []
        for name in layer_names:
            g = f_in[name]
            weight_names = [n.decode('utf8') for n in g.attrs['weight_names']]
            if weight_names:
                filtered_layer_names.append(name)

        layer_names = filtered_layer_names
        if len(layer_names) != len(filtered_layers):
            raise ValueError('You are trying to load a weight file '
                             'containing ' + str(len(layer_names)) +
                             ' layers into a model with ' +
                             str(len(filtered_layers)) + ' layers.')

        # determine permutation vector associated with flattening layer (if it exists)
        flatten_layer_index = -1
        index = 0
        for layer in model.layers:
            if (layer.__class__.__name__.lower() == 'flatten'):
                flatten_layer_index = index
                break
            index = index + 1

        if (flatten_layer_index != -1):
            layer = model.layers[flatten_layer_index]
            permute_layer_name = model.layers[flatten_layer_index + 1].name
            if (image_data_format == 'channels_first'):
                C, H, W = (layer.input_shape)[1:]
            else:
                H, W, C = (layer.input_shape)[1:]
            N = (layer.output_shape)[1]
            perm_index = [0] * N
            if (image_data_format == 'channels_last'):
                ii = 0
                for cc in range(C):
                    for hh in range(H):
                        for ww in range(W):
                            perm_index[ii] = hh * W * C + ww * C + cc
                            ii = ii + 1
            else:
                for nn in range(N):
                    perm_index[nn] = nn
        else:
            perm_index = []

            # let Keras read weights, reformat, and write to SAS-compatible file
        for k, name in enumerate(layer_names):
            g_in = f_in[name]
            g_out = f_out.create_group(name)
            new_weight_names = []

            weight_names = [n.decode('utf8') for n in g_in.attrs['weight_names']]
            weight_values = [g_in[weight_name] for weight_name in weight_names]
            layer = filtered_layers[k]
            symbolic_weights = layer.weights
            weight_values = preprocess_weights_for_loading(layer,
                                                           weight_values,
                                                           original_keras_version,
                                                           original_backend)
            if len(weight_values) != len(symbolic_weights):
                raise ValueError('Layer #' + str(k) +
                                 ' (named "' + layer.name +
                                 '" in the current model) was found to '
                                 'correspond to layer ' + name +
                                 ' in the saved file. '
                                 'However the new layer ' + layer.name +
                                 ' expects ' + str(len(symbolic_weights)) +
                                 ' weights, but the saved weights have ' +
                                 str(len(weight_values)) +
                                 ' elements.')

            # read/write weights
            for ii in range(len(weight_names)):
                tensor_in = np.zeros(weight_values[ii].shape,
                                     dtype=weight_values[ii].dtype)
                weight_values[ii].read_direct(tensor_in)

                # permute axes as needed to conform to SAS deep
                # learning "channels first" format
                if ((image_data_format == 'channels_first') or (not perm_index)):
                    # format: (C,fdim1, fdim2, fdim3) ==> (C,fdim3,fdim1,fdim2)
                    if (len(tensor_in.shape) == 4):
                        tensor_out = np.transpose(tensor_in, (0, 3, 1, 2))
                    else:
                        tensor_out = tensor_in.copy()
                else:
                    # "channels last" format
                    # this is a vector - nothing to permute
                    if (len(tensor_in.shape) == 1):
                        tensor_out = tensor_in.copy()
                    else:
                        # permute Conv2D tensor to "channels_first" format
                        if (layer.__class__.__name__ == 'Conv2D'):
                            tensor_out = np.transpose(tensor_in, (3, 2, 0, 1))
                        # have to account for neuron ordering in first dense
                        # layer following flattening operation
                        elif (layer.__class__.__name__ == 'Dense'):
                            if (layer.name == permute_layer_name):
                                tensor_out = np.zeros(tensor_in.shape)
                                for jj in range(tensor_out.shape[0]):
                                    tensor_out[jj, :] = tensor_in[perm_index[jj], :]
                            else:  # not following flattening, just copy
                                tensor_out = tensor_in.copy()

                            # mimic Caffe layout
                            tensor_out = np.transpose(tensor_out, (1, 0))

                            # save weight in format amenable to SAS
                dset_name = generate_dataset_name(layer, ii)
                new_weight_names.append(dset_name)
                g_out.create_dataset(dset_name, data=tensor_out)

            # update weight names
            g_out.attrs['weight_names'] = new_weight_names

    except ValueError as err_msg:
        print(err_msg)

    finally:
        # close files
        f_out.close()
        f_in.close()

    # generate dataset name with template layerName/layerName/weightName:0


def write_keras_hdf5(model, hdf5_out):
    '''
    Generate an HDF5 file with trained model parameters given a Keras definition

    Parameters
    ----------
    model : Keras model
       Keras deep learning model
    hdf5_out : string
       Fully qualified file name of SAS-compatible HDF5 file

    '''
    # open output file
    try:
        f_out = h5py.File(hdf5_out, 'w')
    except IOError:
        sys.exit('File ' + hdf5_out + ' could not be created')

    try:
        image_data_format = K.image_data_format()

        # determine layers with weights
        filtered_layers = []
        for layer in model.layers:
            weights = layer.weights
            if weights:
                filtered_layers.append(layer)

        # determine permutation vector associated with flattening layer (if it exists)
        flatten_layer_index = -1
        index = 0
        for layer in model.layers:
            if (layer.__class__.__name__.lower() == 'flatten'):
                flatten_layer_index = index
                break
            index = index + 1

        if (flatten_layer_index != -1):
            layer = model.layers[flatten_layer_index]
            permute_layer_name = model.layers[flatten_layer_index + 1].name
            if (image_data_format == 'channels_first'):
                C, H, W = (layer.input_shape)[1:]
            else:
                H, W, C = (layer.input_shape)[1:]
            N = (layer.output_shape)[1]
            perm_index = [0] * N
            if (image_data_format == 'channels_last'):
                ii = 0
                for cc in range(C):
                    for hh in range(H):
                        for ww in range(W):
                            perm_index[ii] = hh * W * C + ww * C + cc
                            ii = ii + 1
            else:
                for nn in range(N):
                    perm_index[nn] = nn
        else:
            perm_index = []

            # let Keras read weights, reformat, and write to SAS-compatible file
        for k, layer in enumerate(filtered_layers):
            # g_in = f_in[name]
            g_out = f_out.create_group(layer.name)
            symbolic_weights = layer.weights
            weight_values = K.batch_get_value(symbolic_weights)
            weight_names = []
            for i, (w, val) in enumerate(zip(symbolic_weights, weight_values)):
                if hasattr(w, 'name') and w.name:
                    name = str(w.name)
                else:
                    name = 'param_' + str(i)
                weight_names.append(name.encode('utf8'))

            # layer modification from here:
            new_weight_names = []

            if len(weight_values) != len(symbolic_weights):
                raise ValueError('Layer #' + str(k) +
                                 ' (named "' + layer.name +
                                 '" in the current model) was found to '
                                 'correspond to layer ' + name +
                                 ' in the saved file. '
                                 'However the new layer ' + layer.name +
                                 ' expects ' + str(len(symbolic_weights)) +
                                 ' weights, but the saved weights have ' +
                                 str(len(weight_values)) +
                                 ' elements.')
            # read/write weights
            for ii in range(len(weight_names)):
                tensor_in = weight_values[ii]

                # permute axes as needed to conform to SAS deep
                # learning "channels first" format
                if ((image_data_format == 'channels_first') or (not perm_index)):
                    # format: (C,fdim1, fdim2, fdim3) ==> (C,fdim3,fdim1,fdim2)
                    if (len(tensor_in.shape) == 4):
                        tensor_out = np.transpose(tensor_in, (0, 3, 1, 2))
                    else:
                        tensor_out = tensor_in.copy()
                else:
                    # "channels last" format
                    # this is a vector - nothing to permute
                    if (len(tensor_in.shape) == 1):
                        tensor_out = tensor_in.copy()
                    else:
                        # permute Conv2D tensor to "channels_first" format
                        if (layer.__class__.__name__ == 'Conv2D'):
                            tensor_out = np.transpose(tensor_in, (3, 2, 0, 1))
                        # have to account for neuron ordering in first dense
                        # layer following flattening operation
                        elif (layer.__class__.__name__ == 'Dense'):
                            if (layer.name == permute_layer_name):
                                tensor_out = np.zeros(tensor_in.shape)
                                for jj in range(tensor_out.shape[0]):
                                    tensor_out[jj, :] = tensor_in[perm_index[jj], :]
                            else:  # not following flattening, just copy
                                tensor_out = tensor_in.copy()

                            # mimic Caffe layout
                            tensor_out = np.transpose(tensor_out, (1, 0))

                            # save weight in format amenable to SAS
                dset_name = generate_dataset_name(layer, ii)
                new_weight_names.append(dset_name)

                g_out.create_dataset(dset_name, data=tensor_out)

            # update weight names
            g_out.attrs['weight_names'] = new_weight_names
    except ValueError as err_msg:
        print(err_msg)

    finally:
        # close files
        f_out.close()
    # generate dataset name with template layerName/layerName/weightName:0


def generate_dataset_name(layer, index):
    '''
    Generate data set names consistent with names generated by Keras models

    Parameters
    ----------
    layer : Layer
       Current layer definition
    index : int
       Data set index

    Returns
    -------
    UTF-8 encoded data set name

    '''
    layer_class_name = layer.__class__.__name__.lower()
    if (layer_class_name in ['conv2d', 'dense']):
        template_names = ['kernel:0', 'bias:0']
    elif (layer_class_name == 'batchnormalization'):
        template_names = ['gamma:0', 'beta:0', 'moving_mean:0', 'moving_variance:0']
    else:
        sys.exit('Unable to translate layer weight name for layer = ' + layer.name)

    dataset_name = layer.name + '/' + template_names[index]
    return dataset_name.encode('utf8')


#########################################################################################
if __name__ == '__main__':
    sys.exit('ERROR: this module cannot be invoked from the command line')



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running CAS server.  You must use an ~/.authinfo
#       file to specify your username and password.  The CAS host and port must
#       be specified using the CASHOST and CASPORT environment variables.
#       A specific protocol ('cas', 'http', 'https', or 'auto') can be set using
#       the CASPROTOCOL environment variable.

import swat
import swat.utils.testing as tm
from dlpy.Sequential import Sequential
from dlpy.layers import InputLayer, Conv2d, Pooling, Dense, OutputLayer

USER, PASSWD = tm.get_user_pass()
HOST, PORT, PROTOCOL = tm.get_host_port_proto()


class TestImageTable(tm.TestCase):
    # Create a class attribute to hold the cas host type
    server_type = None

    def setUp(self):
        swat.reset_option()
        swat.options.cas.print_messages = False
        swat.options.interactive_mode = False

        self.s = swat.CAS(HOST, PORT, USER, PASSWD, protocol=PROTOCOL)

        if type(self).server_type is None:
            # Set once per class and have every test use it. No need to change between tests.
            type(self).server_type = tm.get_cas_host_type(self.s)

        self.srcLib = tm.get_casout_lib(self.server_type)

        # Define the model
        model = Sequential(self.s, model_table='test_model')
        model.add(InputLayer(3, 224, 224, offsets=(0, 0, 0)))
        model.add(Conv2d(8, 7))
        model.add(Pooling(2))
        model.add(Conv2d(8, 7))
        model.add(Pooling(2))
        model.add(Dense(16))
        model.add(OutputLayer(act='softmax', n=2))

        self.model = model

    def tearDown(self):
        # tear down tests
        try:
            self.s.endsession()
        except swat.SWATError:
            pass
        del self.s
        swat.reset_option()

    def test_model_type(self):
        self.assertTrue(isinstance(self.model, Sequential))
        self.assertEqual(self.model.model_table['name'], 'test_model')

    def test_model_info(self):
        out = self.model.get_model_info().ModelInfo
        column_list = ['Model Name', 'Model Type', 'Number of Layers',
                       'Number of Input Layers', 'Number of Output Layers',
                       'Number of Convolutional Layers', 'Number of Pooling Layers',
                       'Number of Fully Connected Layers']
        value_list = ['test_model', 'Convolutional Neural Network', '7', '1', '1', '2', '2', '1']

        for i, v in enumerate(column_list):
            self.assertEqual(out.loc[i].Descr, v)
        for i, v in enumerate(value_list):
            self.assertEqual(out.loc[i].Value.strip(), v)

    def test_layer_types(self):
        layer_type_list = [InputLayer,
                           Conv2d,
                           Pooling,
                           Conv2d,
                           Pooling,
                           Dense,
                           OutputLayer]
        for layer, layer_type in zip(self.model.layers, layer_type_list):
            self.assertTrue(isinstance(layer, layer_type))

    def test_layer_config(self):
        config_keys = [['offsets', 'dropout', 'nchannels', 'width', 'height', 'scale', 'type'],
                       ['act', 'nfilters', 'dropout', 'stride', 'width', 'height', 'type'],
                       ['pool', 'dropout', 'stride', 'width', 'height', 'type'],
                       ['act', 'nfilters', 'dropout', 'stride', 'width', 'height', 'type'],
                       ['pool', 'dropout', 'stride', 'width', 'height', 'type'],
                       ['act', 'dropout', 'n', 'type'],
                       ['act', 'n', 'type']]
        config_values = [[(0, 0, 0), 0, 3, 224, 224, 1, 'input'],
                         ['relu', 8, 0, 1, 7, 7, 'convo'],
                         ['max', 0, 2, 2, 2, 'pool'],
                         ['relu', 8, 0, 1, 7, 7, 'convo'],
                         ['max', 0, 2, 2, 2, 'pool'],
                         ['relu', 0, 16, 'fc'],
                         ['softmax', 2, 'output']]
        for layer, keys, values in zip(self.model.layers, config_keys, config_values):
            for key, value in zip(keys, values):
                self.assertEqual(layer.config[key], value)


if __name__ == '__main__':
    tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' Common layers for the deep learning models '''

from .utils import prod_without_none


class Layer(object):
    '''
    Base class for all layers

    Parameters
    ----------
    name : str
        Specifies the name of the layer.
    config : dict
        Specifies the configuration of the layer.
    src_layers : iter-of-Layers, optional
        Specifies the source layer(s).

    Returns
    -------
    :class:`Layer`

    '''

    def __init__(self, name=None, config=None, src_layers=None):
        self.name = name
        self.config = config
        if isinstance(src_layers, list):
            self.src_layers = src_layers
        elif src_layers is None:
            self.src_layers = None
        else:
            self.src_layers = list(src_layers)

        if 'act' in self.config.keys():
            self.activation = config['act'].title()
        else:
            self.activation = None

        self.output_size = None
        self.kernel_size = None
        self.num_weights = None
        self.num_bias = None

    def to_model_params(self):
        '''
        Convert the model configuration to SAS Viya parameters

        Returns
        -------
        dict

        '''
        if self.config['type'].lower() == 'input':
            return dict(name=self.name, layer=self.config)
        else:
            return dict(name=self.name, layer=self.config,
                        srclayers=[item.name for item in self.src_layers])

    @property
    def summary_str(self):
        '''
        Generate the summary string describing the configuration of the layer.
        '''
        if self.config['type'].lower() == 'input':
            self.output_size = (int(self.config['width']),
                                int(self.config['height']),
                                int(self.config['nchannels']))
            self.kernel_size = None
            self.num_weights = 0
            self.num_bias = 0

        elif self.config['type'].lower() in ('convo', 'convolution'):
            self.output_size = (int(self.src_layers[0].output_size[0] //
                                    self.config['stride']),
                                int(self.src_layers[0].output_size[1] //
                                    self.config['stride']),
                                int(self.config['nfilters']))
            self.kernel_size = (int(self.config['width']), int(self.config['height']))
            self.num_weights = int(self.config['width'] *
                                   self.config['height'] *
                                   self.config['nfilters'] *
                                   self.src_layers[0].output_size[2])
            if 'includeBias' in self.config.keys():
                if self.config['includeBias'] is False:
                    self.num_bias = 0
                else:
                    self.num_bias = int(self.config['nfilters'])
            else:
                self.num_bias = int(self.config['nfilters'])

        elif self.config['type'].lower() in ('pool', 'pooling'):
            self.output_size = (int(self.src_layers[0].output_size[0] //
                                    self.config['stride']),
                                int(self.src_layers[0].output_size[1] //
                                    self.config['stride']),
                                int(self.src_layers[0].output_size[2]))
            self.kernel_size = (int(self.config['width']), int(self.config['height']))
            self.num_weights = 0
            self.num_bias = 0

        elif self.config['type'].lower() == 'batchnorm':
            self.output_size = self.src_layers[0].output_size
            self.kernel_size = None
            self.num_weights = 0
            self.num_bias = int(2 * self.src_layers[0].output_size[2])

        elif self.config['type'].lower() == 'residual':
            self.output_size = (int(min([item.output_size[0]
                                         for item in self.src_layers])),
                                int(min([item.output_size[1]
                                         for item in self.src_layers])),
                                int(max([item.output_size[2]
                                         for item in self.src_layers])))
            self.kernel_size = None
            self.num_weights = 0
            self.num_bias = 0

        elif self.config['type'].lower() == 'concat':
            self.output_size = (int(self.src_layers[0].output_size[0]),
                                int(self.src_layers[0].output_size[1]),
                                int(sum([item.output_size[2]
                                         for item in self.src_layers])))
            self.kernel_size = None
            self.num_weights = 0
            self.num_bias = 0

        elif self.config['type'].lower() in ('fc', 'fullconnect'):
            if isinstance(self.src_layers[0].output_size, int):
                num_features = self.src_layers[0].output_size
            else:
                num_features = prod_without_none(self.src_layers[0].output_size)
            self.output_size = int(self.config['n'])
            self.kernel_size = (int(num_features), int(self.config['n']))
            self.num_weights = int(num_features * self.config['n'])
            self.num_bias = int(self.config['n'])

        elif self.config['type'].lower() == 'output':
            self.type_name = 'Output'
            if isinstance(self.src_layers[0].output_size, int):
                num_features = self.src_layers[0].output_size
            else:
                num_features = prod_without_none(self.src_layers[0].output_size)

            if 'n' in self.config.keys():
                self.output_size = int(self.config['n'])
                self.kernel_size = (int(num_features), int(self.config['n']))
                self.num_weights = int(num_features * self.config['n'])
                self.num_bias = int(self.config['n'])
            else:
                self.kernel_size = None
                self.num_weights = None
                self.num_bias = None
                self.output_size = None

        name = '{}({})'.format(self.name, self.type_name)
        if len(name) > 17:
            col1 = '| {:<17}'.format('{}'.format(name[:14] + '...'))
        else:
            col1 = '| {:<17}'.format('{}'.format(name))

        col2 = '|{:^15}'.format('{}'.format(self.kernel_size))

        if 'stride' not in self.config.keys():
            col3 = '|{:^8}'.format('None')
        else:
            col3 = '|{:^8}'.format('{}'.format(int(self.config['stride'])))

        col4 = '|{:^12}'.format('{}'.format(self.activation))

        col5 = '|{:^17}'.format('{}'.format(self.output_size))

        num_paras = '{} / {}'.format(self.num_weights, self.num_bias)
        col6 = '|{:^22}|\n'.format(num_paras)

        return col1 + col2 + col3 + col4 + col5 + col6


class InputLayer(Layer):
    '''
    Input layer
    
    Parameters
    ----------
    n_channels : int
        Specifies the number of channels of the input images.
    width : int
        Specifies the width of the input images.
    height : int
        Specifies the height of the input images.
    scale : double, between 0 and 1.
        Specifies the scaling parameter of the image.
    dropout : double, between 0 and 1
        Specifies the dropout rate.
    offsets : iter-of-doubles
        Specifies the offset values.
    name : str
        Specifies the name of the layer.
    src_layers : iter-of-Layers, optional
        Specifies the source layer(s).

    Returns
    -------
    :class:`InputLayer`

    '''

    def __init__(self, n_channels=3, width=224, height=224, scale=1,
                 dropout=0, offsets=None, name=None, **kwargs):
        if offsets is None:
            offsets = [0] * n_channels
        config = locals()
        config = _unpack_config(config)
        config['type'] = 'input'
        Layer.__init__(self, name, config)
        self.color_code = '#F0FF00'
        self.type_name = 'Input'


class Conv2d(Layer):
    '''
    2D convolutional layer

    Parameters
    ----------
    n_filters : int
        Specifies the number of filters.
    width : int
        Specifies the width of pooling window.
    height : int
        Specifies the height of pooling window.
    stride : int
        Specifies the step size of the moving window.
    act : str
        Specifies the activation types.
    dropout : double, between 0 and 1
        Specifies the dropout rate.
    name : str
        Specifies the name of the layer.
    src_layers : iter-of-Layers, optional
        Specifies the source layer(s).

    Returns
    -------
    :class:`Conv2d`

    '''

    def __init__(self, n_filters, width=None, height=None, stride=1,
                 act='relu', dropout=0, name=None, src_layers=None, **kwargs):
        if (width is None) and (height is None):
            width = 3
        if width is None:
            width = height
        if height is None:
            height = width
        config = locals()
        config = _unpack_config(config)
        config['type'] = 'convo'
        Layer.__init__(self, name, config, src_layers)
        self.color_code = '#6CFF00'
        self.type_name = 'Convo.'


class Pooling(Layer):
    '''
    Pooling layer

    Parameters
    ----------
    width : int
        Specifies the width of pooling window.
    height : int
        Specifies the height of pooling window.
    stride : int
        Specifies the step size of the moving window.
    act : str
        Specifies the activation types.
    dropout : double, between 0 and 1
        Specifies the dropout rate.
    name : str
        Specifies the name of the layer.
    src_layers : iter-of-Layers, optional
        Specifies the source layer(s).

    Returns
    -------
    :class:`Pooling`

    '''

    def __init__(self, width=None, height=None, stride=None,
                 pool='max', dropout=0, name=None, src_layers=None, **kwargs):
        if (width is None) and (height is None):
            width = 2
        if width is None:
            width = height
        if height is None:
            height = width
        if stride is None:
            stride = width
        config = locals()
        config = _unpack_config(config)
        config['type'] = 'pool'
        Layer.__init__(self, name, config, src_layers)
        self.activation = pool.title()
        self.color_code = '#FF9700'
        self.type_name = 'Pool'


class Dense(Layer):
    '''
    Fully connected layer

    Parameters
    ----------
    n : int
        Specifies the number of neurons.
    act : str
        Specifies the activation types.
    dropout : double, between 0 and 1
        Specifies the dropout rate.
    name : str
        Specifies the name of the layer.
    src_layers : iter-of-Layers, optional
        Specifies the source layer(s).

    Returns
    -------
    :class:`Dense`

    '''

    def __init__(self, n, act='relu', dropout=0,
                 name=None, src_layers=None, **kwargs):
        config = locals()
        config = _unpack_config(config)
        config['type'] = 'fc'
        Layer.__init__(self, name, config, src_layers)
        self.color_code = '#00ECFF'
        self.type_name = 'F.C.'


class Recurrent(Layer):
    '''
    Recurrent layer

    Parameters
    ----------
    n : int
        Specifies the number of neurons.
    act : str
        Specifies the activation types.
    rnn_type : str
        Specifies the type of RNN gate.
    output_type : str
        Specifies the type of output neurons.
    reversed_ : bool
        Specifies whether to reverse the sequence.
    name : str
        Specifies the name of the layer.
    src_layers : iter-of-Layers, optional
        Specifies the source layer(s).

    Returns
    -------
    :class:`Recurrent`

    '''

    def __init__(self, n, act='AUTO', rnn_type='RNN', output_type='ENCODING',
                 reversed_=False, name=None, src_layers=None, **kwargs):
        config = locals()
        config = _unpack_config(config)
        config['type'] = 'recurrent'
        Layer.__init__(self, name, config, src_layers)
        self.color_code = '#FFA4A4'
        self.type_name = 'Rec.'


class BN(Layer):
    '''
    Batch Normalization layer

    Parameters
    ----------
    act : str
        Specifies the activation types.
    name : str
        Specifies the name of the layer.
    src_layers : iter-of-Layers, optional
        Specifies the source layer(s).

    Returns
    -------
    :class:`BN`

    '''

    def __init__(self, act='AUTO', name=None, src_layers=None, **kwargs):
        config = locals()
        config = _unpack_config(config)
        config['type'] = 'batchnorm'
        Layer.__init__(self, name, config, src_layers)
        self.color_code = '#FFF999'
        self.type_name = 'B.N.'


class Res(Layer):
    '''
    Residual layer

    Parameters
    ----------
    act : str
        Specifies the activation types.
    name : str
        Specifies the name of the layer.
    src_layers : iter-of-Layers, optional.
        Specifies the source layer(s).

    Returns
    -------
    :class:`Res`

    '''

    def __init__(self, act='AUTO', name=None, src_layers=None, **kwargs):
        config = locals()
        config = _unpack_config(config)
        config['type'] = 'residual'
        Layer.__init__(self, name, config, src_layers)
        self.color_code = '#FF0000'
        self.type_name = 'Resid.'


class Concat(Layer):
    '''
    Concatenation layer

    Parameters
    ----------
    act : str
        Specifies the activation types.
    name : str
        Specifies the name of the layer.
    src_layers : iter-of-Layers, optional
        Specifies the source layer(s).

    Returns
    -------
    :class:`Concat`

    '''

    def __init__(self, act='AUTO', name=None, src_layers=None, **kwargs):
        config = locals()
        config = _unpack_config(config)
        config['type'] = 'concat'
        Layer.__init__(self, name, config, src_layers)
        self.color_code = '#DD5022'
        self.type_name = 'Concat.'


class Proj(Layer):
    '''
    Projection layer

    Parameters
    ----------
    name : str
        Specifies the name of the layer.
    src_layers : iter-of-Layers, optional
        Specifies the source layer(s).

    Returns
    -------
    :class:`Proj`

    '''

    def __init__(self, name=None, src_layers=None, **kwargs):
        config = locals()
        config = _unpack_config(config)
        config['type'] = 'projection'
        Layer.__init__(self, name, config, src_layers)
        self.color_code = '#FFA2A3'
        self.type_name = 'Proj.'


class OutputLayer(Layer):
    '''
    Output layer

    Parameters
    ----------
    n : int
        Specifies the number of output neurons.
    act : str
        Specifies the activation types.
    name : str
        Specifies the name of the layer.
    src_layers : iter-of-Layers, optional
        Specifies the source layer(s).

    Returns
    -------
    :class:`OutputLayer`
 
    '''

    def __init__(self, n=None, act='softmax', name=None, src_layers=None, **kwargs):
        config = locals()
        config = _unpack_config(config)
        config['type'] = 'output'
        if config['n'] is None:
            del config['n']
        Layer.__init__(self, name, config, src_layers)
        self.color_code = '#C8C8C8'
        self.type_name = 'Output.'


def _unpack_config(config):
    ''' Unpack the configuration from the keyword-argument-only input '''
    kwargs = config['kwargs']
    del config['self'], config['name'], config['kwargs']
    try:
        del config['src_layers']
    except:
        pass
    out = {}
    out.update(config)
    out.update(kwargs)
    for key in out:
        if '_' in key:
            new_key = key.replace('_', '')
            out[new_key] = out[key]
            out.pop(key)
    return out



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' LeNet with batch normalization model definition '''


def LeNet_Model(s, model_name='LeNet'):
    '''
    LeNet model definition (batch normalization version)

    Parameters
    ----------
    s : CAS
        Specifies the CAS connection object
    model_name : string, optional
        Specifies the name of CAS table to store the model
    include_top : boolean, optional
        Specifies whether to include the top layers of the model.

    Returns
    -------
    None
        A CAS table defining the model is created

    '''

    # instantiate model
    s.deepLearn.buildModel(model=dict(name=model_name, replace=True), type='CNN')

    # input layer
    s.deepLearn.addLayer(model=model_name, name='mnist',
                         layer=dict(type='input', n_channels=1, width=28, height=28,
                                    scale=0.00392156862745098039))

    # conv1: 5*5*20
    s.deepLearn.addLayer(model=model_name, name='conv1',
                         layer=dict(type='convolution', nFilters=20, width=5, height=5,
                                    stride=1, act='identity', noBias=True, init='xavier'),
                         srcLayers=['mnist'])

    # conv1 batch normalization
    s.deepLearn.addLayer(model=model_name, name='conv1_bn',
                         layer=dict(type='batchnorm', act='relu'), srcLayers=['conv1'])

    # pool1 2*2*2
    s.deepLearn.addLayer(model=model_name, name='pool1',
                         layer=dict(type='pooling', width=2, height=2, stride=2, pool='max'),
                         srcLayers=['conv1_bn'])

    # conv2: 5*5*50
    s.deepLearn.addLayer(model=model_name, name='conv2',
                         layer=dict(type='convolution', nFilters=50, width=5, height=5,
                                    stride=1, act='identity', noBias=True, init='xavier'),
                         srcLayers=['pool1'])

    # conv2 batch normalization
    s.deepLearn.addLayer(model=model_name, name='conv2_bn',
                         layer=dict(type='batchnorm', act='relu'), srcLayers=['conv2'])

    # pool2 2*2*2
    s.deepLearn.addLayer(model=model_name, name='pool2',
                         layer=dict(type='pooling', width=2, height=2, stride=2, pool='max'),
                         srcLayers=['conv2_bn'])

    # fully connected layer
    s.deepLearn.addLayer(model=model_name, name='ip1',
                         layer=dict(type='fullconnect', n=500, init='xavier', act='relu'),
                         srcLayers=['pool2'])
    # output layer
    s.deepLearn.addLayer(model=model_name, name='ip2',
                         layer=dict(type='output', n=10, init='xavier', act='softmax'),
                         srcLayers=['ip1'])



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from ..utils import input_table_check


def VGG19_Model(s, model_table='VGG19', n_channels=3, width=224, height=224,
                random_crop=None, offsets=None):
    '''
    VGG19 model definition

    Parameters
    ----------
    s : CAS
        Specifies the CAS connection object
    model_table : string, dict or CAS table, optional
        Specifies the CAS table to store the model.
    n_channels : int, optional
        Specifies the number of the channels of the input layer
        Default: 3
    width : int, optional
        Specifies the width of the input layer
        Default: 224
    height : int, optional
        Specifies the height of the input layer
        Default: 224
    random_crop : string, optional
        Specifies how to crop the data in the input layer when image data is
        used. Images are cropped to the values that are specified in the width
        and height parameters.deepLearn. Only the images with one or both dimensions
        that are larger than those sizes are cropped.
        Valid Values: 'none' or 'unique'
        Default	: 'unique'
    offsets : double or iter-of-doubles, optional
        Specifies an offset for each channel in the input data. The final
        input data is set after applying scaling and subtracting the
        specified offsets.deepLearn.
        Default: (103.939, 116.779, 123.68)

    Returns
    -------
    None
        A CAS table defining the model is created

    '''
    model_table_opts = input_table_check(model_table)

    if random_crop is None:
        random_crop = 'none'
    elif random_crop.lower() not in ['none', 'unique']:
        raise ValueError('random_crop can only be "none" or "unique"')

    if offsets is None:
        offsets = [103.939, 116.779, 123.68]

    # instantiate model
    s.deepLearn.buildModel(model=dict(replace=True, **model_table_opts), type='CNN')

    # input layer
    s.deepLearn.addLayer(model=model_table, name='data',
                         layer=dict(type='input', nchannels=n_channels, width=width, height=height,
                                    randomcrop=random_crop, offsets=offsets))

    # conv1_1 layer: 64*3*3
    s.deepLearn.addLayer(model=model_table, name='conv1_1',
                         layer=dict(type='convolution', nFilters=64, width=3, height=3,
                                    stride=1, act='relu'),
                         srcLayers=['data'])

    # conv1_2 layer: 64*3*3
    s.deepLearn.addLayer(model=model_table, name='conv1_2',
                         layer=dict(type='convolution', nFilters=64, width=3, height=3,
                                    stride=1, act='relu'),
                         srcLayers=['conv1_1'])

    # pool1 layer: 2*2
    s.deepLearn.addLayer(model=model_table, name='pool1',
                         layer=dict(type='pooling', width=2, height=2, stride=2, pool='max'),
                         srcLayers=['conv1_2'])

    # conv2_1 layer: 128*3*3
    s.deepLearn.addLayer(model=model_table, name='conv2_1',
                         layer=dict(type='convolution', nFilters=128, width=3, height=3,
                                    stride=1, act='relu'),
                         srcLayers=['pool1'])

    # conv2_2 layer: 128*3*3
    s.deepLearn.addLayer(model=model_table, name='conv2_2',
                         layer=dict(type='convolution', nFilters=128, width=3, height=3,
                                    stride=1, act='relu'),
                         srcLayers=['conv2_1'])

    # pool2 layer: 2*2
    s.deepLearn.addLayer(model=model_table, name='pool2',
                         layer=dict(type='pooling', width=2, height=2, stride=2, pool='max'),
                         srcLayers=['conv2_2'])

    # conv3_1 layer: 256*3*3
    s.deepLearn.addLayer(model=model_table, name='conv3_1',
                         layer=dict(type='convolution', nFilters=256, width=3, height=3,
                                    stride=1, act='relu'),
                         srcLayers=['pool2'])

    # conv3_2 layer: 256*3*3
    s.deepLearn.addLayer(model=model_table, name='conv3_2',
                         layer=dict(type='convolution', nFilters=256, width=3, height=3,
                                    stride=1, act='relu'),
                         srcLayers=['conv3_1'])

    # conv3_3 layer: 256*3*3
    s.deepLearn.addLayer(model=model_table, name='conv3_3',
                         layer=dict(type='convolution', nFilters=256, width=3, height=3,
                                    stride=1, act='relu'),
                         srcLayers=['conv3_2'])

    # conv3_4 layer: 256*3*3
    s.deepLearn.addLayer(model=model_table, name='conv3_4',
                         layer=dict(type='convolution', nFilters=256, width=3, height=3,
                                    stride=1, act='relu'),
                         srcLayers=['conv3_3'])

    # pool3 layer: 2*2
    s.deepLearn.addLayer(model=model_table, name='pool3',
                         layer=dict(type='pooling', width=2, height=2, stride=2, pool='max'),
                         srcLayers=['conv3_4'])

    # conv4_1 layer: 512*3*3
    s.deepLearn.addLayer(model=model_table, name='conv4_1',
                         layer=dict(type='convolution', nFilters=512, width=3, height=3,
                                    stride=1, act='relu'),
                         srcLayers=['pool3'])

    # conv4_2 layer: 512*3*3
    s.deepLearn.addLayer(model=model_table, name='conv4_2',
                         layer=dict(type='convolution', nFilters=512, width=3, height=3,
                                    stride=1, act='relu'),
                         srcLayers=['conv4_1'])

    # conv4_3 layer: 512*3*3
    s.deepLearn.addLayer(model=model_table, name='conv4_3',
                         layer=dict(type='convolution', nFilters=512, width=3, height=3,
                                    stride=1, act='relu'),
                         srcLayers=['conv4_2'])

    # conv4_4 layer: 512*3*3 */
    s.deepLearn.addLayer(model=model_table, name='conv4_4',
                         layer=dict(type='convolution', nFilters=512, width=3, height=3,
                                    stride=1, act='relu'),
                         srcLayers=['conv4_3'])

    # pool4 layer: 2*2
    s.deepLearn.addLayer(model=model_table, name='pool4',
                         layer=dict(type='pooling', width=2, height=2, stride=2, pool='max'),
                         srcLayers=['conv4_4'])

    # conv5_1 layer: 512*3*3
    s.deepLearn.addLayer(model=model_table, name='conv5_1',
                         layer=dict(type='convolution', nFilters=512, width=3, height=3,
                                    stride=1, act='relu'),
                         srcLayers=['pool4'])

    # conv5_2 layer: 512*3*3
    s.deepLearn.addLayer(model=model_table, name='conv5_2',
                         layer=dict(type='convolution', nFilters=512, width=3, height=3,
                                    stride=1, act='relu'),
                         srcLayers=['conv5_1'])

    # conv5_3 layer: 512*3*3
    s.deepLearn.addLayer(model=model_table, name='conv5_3',
                         layer=dict(type='convolution', nFilters=512, width=3, height=3,
                                    stride=1, act='relu'),
                         srcLayers=['conv5_2'])

    # conv5_4 layer: 512*3*3
    s.deepLearn.addLayer(model=model_table, name='conv5_4',
                         layer=dict(type='convolution', nFilters=512, width=3, height=3,
                                    stride=1, act='relu'),
                         srcLayers=['conv5_3'])

    # pool5 layer: 2*2
    s.deepLearn.addLayer(model=model_table, name='pool5',
                         layer=dict(type='pooling', width=2, height=2, stride=2, pool='max'),
                         srcLayers=['conv5_4'])

    # fc6 layer: 4096 neurons
    s.deepLearn.addLayer(model=model_table, name='fc6',
                         layer=dict(type='fullconnect', n=4096, act='relu', dropout=0.5),
                         srcLayers=['pool5'])

    # fc7 layer: 4096 neurons
    s.deepLearn.addLayer(model=model_table, name='fc7',
                         layer=dict(type='fullconnect', n=4096, act='relu', dropout=0.5),
                         srcLayers=['fc6'])
    # fc output layer: 1000 neurons
    s.deepLearn.addLayer(model=model_table, name='fc8',
                         layer=dict(type='output', n=1000, act='softmax'),
                         srcLayers=['fc7'])

    return s.CASTable(**model_table_opts)
